// iPaq 3D engine  -  Janurary 2001  -  The Phantom
// Minimal 3D engine for Handhelds
// ------------------------------------------------

#include "easyce.h"
#ifdef _WIN32_WCE
#include "mmsystem.h"
#endif
#include "engine.h"
#include "rasterize.h"
#include "math.h"
#include "stdio.h"
#include "string.h"
#include "stdlib.h"

// ------------------------------------------------
// AVector::dot
// Dotproduct
// ------------------------------------------------
int AVector::dot( AVector a_Second )
{
	return ( FPMUL( m_Cell[0], a_Second.m_Cell[0] ) + 
			 FPMUL( m_Cell[1], a_Second.m_Cell[1] ) +
			 FPMUL( m_Cell[2], a_Second.m_Cell[2] ) );
}

// ------------------------------------------------
// AVector::cross
// Crossproduct
// ------------------------------------------------
AVector AVector::cross( AVector a_Second )
{
	return AVector(
		FPMUL( m_Cell[1], a_Second.m_Cell[2] ) - FPMUL( m_Cell[2], a_Second.m_Cell[1] ),
		FPMUL( m_Cell[2], a_Second.m_Cell[0] ) - FPMUL( m_Cell[0], a_Second.m_Cell[2] ),
		FPMUL( m_Cell[0], a_Second.m_Cell[1] ) - FPMUL( m_Cell[1], a_Second.m_Cell[0] ) );
}

// ------------------------------------------------
// AVector::length
// Vector length
// ------------------------------------------------
int AVector::length()
{
	return 0;
}

// ------------------------------------------------
// AVector::normalize
// Normalize vector
// ------------------------------------------------
void AVector::normalize()
{
}

// ------------------------------------------------
// AMatrix::identity
// Build the identity matrix
// ------------------------------------------------
void AMatrix::identity()
{
	for ( int i = 0; i < 16; i++ ) m_Cell[i] = 0;
	m_Cell[0] = m_Cell[5] = m_Cell[10] = m_Cell[15] = FP_ONE;
}

// ------------------------------------------------
// AMatrix::rotatex
// Create a rotational matrix for x only
// ------------------------------------------------
void AMatrix::rotatex( int a_X )
{
	int sinx = FP( sin( (float)a_X / FP_ONE * PI / 128.0f ) );
	int cosx = FP( cos( (float)a_X / FP_ONE * PI / 128.0f ) );
	identity();
	m_Cell[5] = cosx;
	m_Cell[6] = sinx;
	m_Cell[9] = -sinx;
	m_Cell[10] = cosx;
}

// ------------------------------------------------
// AMatrix::rotatey
// Create a rotational matrix for y only
// ------------------------------------------------
void AMatrix::rotatey( int a_Y )
{
	int siny = FP( sin( (float)a_Y / FP_ONE * PI / 128.0f ) );
	int cosy = FP( cos( (float)a_Y / FP_ONE * PI / 128.0f ) );
	identity();
	m_Cell[0] = cosy;
	m_Cell[2] = -siny;
	m_Cell[8] = siny;
	m_Cell[10] = cosy;
}

// ------------------------------------------------
// AMatrix::rotatey
// Create a rotational matrix for z only
// ------------------------------------------------
void AMatrix::rotatez( int a_Z )
{
	int sinz = FP( sin( (float)a_Z / FP_ONE * PI / 128.0f ) );
	int cosz = FP( cos( (float)a_Z / FP_ONE * PI / 128.0f ) );
	identity();
	m_Cell[0] = cosz;
	m_Cell[1] = sinz;
	m_Cell[4] = -sinz;
	m_Cell[5] = cosz;
}

// ------------------------------------------------
// AMatrix::translate
// Set the translational part of the matrix
// ------------------------------------------------
void AMatrix::translate( int a_X, int a_Y, int a_Z )
{
	m_Cell[3] += a_X;
	m_Cell[7] += a_Y;
	m_Cell[11] += a_Z;
}

// ------------------------------------------------
// AMatrix::concat
// Concatenate two matrices
// ------------------------------------------------
void AMatrix::concat( const AMatrix& a_M2)
{
	AMatrix res;
	for ( int col = 0; col < 4; col++ )
		for ( int row = 0; row < 4; row++ )
			res.m_Cell[row * 4 + col] = 
				FPMUL( m_Cell[row * 4], a_M2.m_Cell[col] ) +
				FPMUL( m_Cell[row * 4 + 1], a_M2.m_Cell[col + 4] ) +
				FPMUL( m_Cell[row * 4 + 2], a_M2.m_Cell[col + 8] );
	for ( col = 0; col < 16; col++ ) m_Cell[col] = res.m_Cell[col];
}

// ------------------------------------------------
// AMatrix::rotate
// Create a rotational matrix
// ------------------------------------------------
void AMatrix::rotate( int a_RX, int a_RY, int a_RZ )
{	
	AMatrix tempmat;
	tempmat.rotatex( a_RX );
	this->rotatey( a_RY );
	this->concat( tempmat );
	tempmat.rotatez( a_RZ );
	this->concat( tempmat );
}

// ------------------------------------------------
// AMatrix::getz
// Return the z vector of this matrix
// ------------------------------------------------
AVector AMatrix::getz()
{
	return AVector( m_Cell[2], m_Cell[6], m_Cell[10] );
}

// ------------------------------------------------
// ACoord::transform
// Transform the coordinate using the specified matrix
// ------------------------------------------------
void ACoord::transform( AMatrix a_Mat )
{
	int ox = m_Orig.getx(), oy = m_Orig.gety(), oz = m_Orig.getz();
	m_Rot.setxyz( FPMUL( a_Mat.get( 0 ), ox ) + 
				  FPMUL( a_Mat.get( 1 ), oy ) + 
				  FPMUL( a_Mat.get( 2 ), oz ) + a_Mat.get( 3 ),
				  FPMUL( a_Mat.get( 4 ), ox ) + 
				  FPMUL( a_Mat.get( 5 ), oy ) + 
				  FPMUL( a_Mat.get( 6 ), oz ) + a_Mat.get( 7 ),
				  FPMUL( a_Mat.get( 8 ), ox ) + 
				  FPMUL( a_Mat.get( 9 ), oy ) + 
				  FPMUL( a_Mat.get( 10 ), oz ) + a_Mat.get( 11 ) );
}

// ------------------------------------------------
// ACoord::perspective
// Determine the final screen coordinate
// ------------------------------------------------
void ACoord::perspective()
{
	m_SX = IFP(160) + FPDIV( m_Rot.getx(), m_Rot.getz() >> 9 );
	m_SY = IFP(120) + FPDIV( m_Rot.gety(), m_Rot.getz() >> 9 );
}

// ------------------------------------------------
// APolygon::APolygon
// Constructor
// ------------------------------------------------
APolygon::APolygon()
{
	for ( int i = 0; i < 3; i++ ) m_Vert[i] = 0;
	m_Text = 0;
}

// ------------------------------------------------
// APolygon::~APolygon
// Destructor
// ------------------------------------------------
APolygon::~APolygon()
{
}

// ------------------------------------------------
// APolygon::facing
// 2D clockwise test
// ------------------------------------------------
bool APolygon::facing()
{
	ACoord* c1 = m_Vert[0]->getcoord();
	ACoord* c2 = m_Vert[1]->getcoord();
	ACoord* c3 = m_Vert[2]->getcoord();
	int sx1 = c1->getscreenx() >> (FPP - 6);
	int sy1 = c1->getscreeny() >> (FPP - 6);
	int sx2 = c2->getscreenx() >> (FPP - 6);
	int sy2 = c2->getscreeny() >> (FPP - 6);
	int sx3 = c3->getscreenx() >> (FPP - 6);
	int sy3 = c3->getscreeny() >> (FPP - 6);
	return ((((sx1 - sx2) * (sy3 - sy2)) - 
	         ((sy1 - sy2) * (sx3 - sx2))) > 0);
}

// ------------------------------------------------
// APolygon::abcd
// Calc plane equation for polygon
// ------------------------------------------------
void APolygon::abcd()
{
	AVector v1 = m_Vert[0]->getcoord()->getorig();
	AVector v2 = m_Vert[1]->getcoord()->getorig();
	AVector v3 = m_Vert[2]->getcoord()->getorig();
	double rx1 = (double)(v1.getx() - v1.getx()) / FP_ONE;
	double ry1 = (double)(v2.gety() - v1.gety()) / FP_ONE;
	double rz1 = (double)(v2.getz() - v1.getz()) / FP_ONE;
	double rx2 = (double)(v3.getx() - v1.getx()) / FP_ONE;
	double ry2 = (double)(v3.gety() - v1.gety()) / FP_ONE;
	double rz2 = (double)(v3.getz() - v1.getz()) / FP_ONE;
	double A = (ry1 * rz2) - (ry2 * rz1 );
	double B = (rz1 * rx2) - (rz2 * rx1 );
	double C = (rx1 * ry2) - (rx2 * ry1 );
	double len = sqrt( A * A + B * B + C * C);
	m_A = (int)(A * FP_ONE / len);
	m_B = (int)(B * FP_ONE / len);
	m_C = (int)(C * FP_ONE / len);
	m_D = (int)(A * v2.getx() + B * v2.gety() + C * v2.getz());
}

// ------------------------------------------------
// AMesh::AMesh
// Constructor
// ------------------------------------------------
AMesh::AMesh()
{
	m_Crd = 0;
	m_Poly = 0;
	m_Coords = m_Polies = 0;
}

// ------------------------------------------------
// AMesh::~AMesh
// Destructor
// ------------------------------------------------
AMesh::~AMesh()
{
	delete m_Crd;
	delete m_Poly;
}

// ------------------------------------------------
// AMesh::transform
// Transform the mesh using the given matrix
// ------------------------------------------------
void AMesh::transform( AMatrix a_Mat )
{
	for ( int i = 0; i < m_Coords; i++ ) 
	{
		m_Crd[i]->transform( a_Mat );
		m_Crd[i]->perspective();
	}
}

// ------------------------------------------------
// AMesh::draw
// Draw the mesh
// ------------------------------------------------
void AMesh::draw( ARasterizer* a_Rast, AVector a_Dir )
{
	for ( int i = 0; i < m_Polies; i++ ) 
	{
		if (m_Poly[i]->facing()) a_Rast->draw( m_Poly[i] );
	}
}

// ------------------------------------------------
// AMesh::init
// Setup the internal arrays of the mesh
// ------------------------------------------------
void AMesh::init( int a_Polies, int a_Crds )
{
	m_Crd = new ACoord*[a_Crds];
	m_Poly = new APolygon*[a_Polies];
	m_Coords = 0;
	m_Polies = 0;
}

// ------------------------------------------------
// AMesh::load3ds
// Basic 3ds loader
// ------------------------------------------------
#ifdef _WIN32_WCE
void AMesh::load3ds( unsigned short* filename )
#else
void AMesh::load3ds( char* filename )
#endif
{
	HANDLE file;
	bool doneloading = false;
	unsigned long filelength;
	char* memfile;
	file = CreateFile( filename, GENERIC_READ, FILE_SHARE_READ,
					   0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0 );
	if (file == INVALID_HANDLE_VALUE) return;
	memfile = new char[800 * 1024];
	ReadFile( file, memfile, 800 * 1024, &filelength, 0 );
	CloseHandle( file );
	eatchunk( memfile );
	delete memfile;
#ifdef _WIN32_WCE
	ATexture* text = new ATexture( L"\\Program Files\\EasyCE\\cowskin.tga", 256, 256 );
#else
	ATexture* text = new ATexture( "cowskin.tga", 256, 256 );
#endif
	int xmax = IFP(-1000), xmin = IFP(1000);
	int ymax = IFP(-1000), ymin = IFP(1000);
	int zmax = IFP(-1000), zmin = IFP(1000);
	for ( int i = 0; i < m_Coords; i++ )
	{
		int x = m_Crd[i]->getorig().getx();
		int y = m_Crd[i]->getorig().gety();
		int z = m_Crd[i]->getorig().getz();
		if (x < xmin) xmin = x;
		if (x > xmax) xmax = x;
		if (y < ymin) ymin = y;
		if (y > ymax) ymax = y;
		if (z < zmin) zmin = z;
		if (z > zmax) zmax = z;
	}
	double du = (FP_ONE * 256.0) / (double)(xmax - xmin);
	double dv = (FP_ONE * 256.0) / (double)(zmax - zmin);
	for ( i = 0; i < m_Polies; i++ )
	{
		m_Poly[i]->settexture( text );
		for ( int j = 0; j < 3; j++ )
		{
			AVertex* v = m_Poly[i]->getvertex( j );
			AVector p = v->getcoord()->getorig();
			v->setuv1( (int)((p.getx() - xmin) * du), 
					   (int)((p.getz() - zmin) * dv) );
		}
	}
	return;
}

// Code to make reads from odd addresses safe
unsigned char* temp = new unsigned char[8];
unsigned char* aligned = (unsigned char*)(((unsigned long)temp + 4) & 0xFFFFFFFC);
inline float getf( char* addr )
{
	memcpy( aligned, addr, 4 );
	return *(float*)aligned;
}
inline unsigned short getw( char* addr )
{
	memcpy( aligned, addr, 2 );
	return *(unsigned short*)aligned;
}
inline unsigned long getd( char* addr )
{
	memcpy( aligned, addr, 4 );
	return *(unsigned long*)aligned;
}

// ------------------------------------------------
// AMesh::eatchunk
// Recursive 3ds chunk reader
// ------------------------------------------------
long AMesh::eatchunk( char* buffer )
{
	short chunkid = getw(buffer);
	long chunklength = getd(buffer+2);
	int j, i = 6;
	switch (chunkid)
	{
	case 0x4D4D:
		while ((getw(buffer+i) != 0x3D3D) &&
			(getw(buffer+i) != 0xB000)) i += 2;
		break;
	case 0x3D3D:
		break;
	case 0x4000:
		while (*(buffer+(i++)) != 0);
		break;
	case 0x4100:
		break;
	case 0x4110:
		m_Coords = getw(buffer+i); 	
		i+=2;
		m_Crd = new ACoord*[m_Coords];
		for ( j = 0; j < m_Coords; j++ )
		{
			int x = (int)((FP_ONE >> 3) * getf(buffer+i));
			i+=4;
			int y = (int)((FP_ONE >> 3) * getf(buffer+i));
			i+=4;
			int z = (int)((FP_ONE >> 3) * getf(buffer+i));
			i+=4;
			m_Crd[j] = new ACoord( x, y, z );
		}
		break;
	case 0x4120:
		{		
			int cp = 0;
			m_Polies = getw(buffer+i);
			i+=2;
			m_Poly = new APolygon*[m_Polies];
			for ( j = 0; j < m_Polies; j++ )
			{
				int c1 = getw(buffer+i);
				i+=2;
				int c2 = getw(buffer+i);
				i+=2;
				int c3 = getw(buffer+i);
				i+=2;
				i+=2;
				if ((c1 != c2) && (c2 != c3) && (c1 != c3))
				{
					AVertex* v1 = new AVertex( m_Crd[c1], 0, 0 );
					AVertex* v2 = new AVertex( m_Crd[c2], 0, 0 );
					AVertex* v3 = new AVertex( m_Crd[c3], 0, 0 );
					m_Poly[cp] = new APolygon( v1, v2, v3 );
					m_Poly[cp++]->abcd();
				}
			}
		}
		break;
	default:
		i = chunklength;
	break;
	}
	while (i < chunklength) i += eatchunk( buffer + i );
	return chunklength;
}

// ------------------------------------------------
// ATexture::ATexture
// Constructor that also creates a nice image buffer
// ------------------------------------------------
ATexture::ATexture( int a_Width, int a_Height )
{
	m_Width = a_Width;
	m_Height = a_Height;
	m_Bitmap = new PIXEL[m_Width * m_Height];
}

// ------------------------------------------------
// ATexture::loadtga
// Loads a tga file to a texture
// ------------------------------------------------
#ifdef _WIN32_WCE
void ATexture::loadtga( unsigned short* a_File, int a_Width, int a_Height )
#else
void ATexture::loadtga( char* a_File, int a_Width, int a_Height )
#endif
{
	m_Width = a_Width;
	m_Height = a_Height;
	delete m_Bitmap;
	m_Bitmap = new PIXEL[m_Width * m_Height];
	PIXEL* temp = new PIXEL[m_Width * m_Height + 9];
#ifdef _WIN32_WCE
	HANDLE tga = CreateFile( a_File, GENERIC_READ, FILE_SHARE_READ,
	                         0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0 );
	if (tga == INVALID_HANDLE_VALUE) return;
	unsigned long bytesread;
	ReadFile( tga, temp, a_Width * a_Height * 2 + 18, &bytesread, 0 );
	if (CloseHandle( tga ) == 0) return;
#else
	FILE* tga = fopen( a_File, "rb" );
	fread( temp, m_Width * m_Height + 9, 2, tga );
	fclose( tga );
#endif
	for ( int i = 0; i < a_Width * a_Height; i++ )
	{
		PIXEL src = *(temp + i + 9);
		int red = (src >> 11) & 31;
		int green = (src >> 5) & 31;
		int blue = src & 31;
		*(m_Bitmap + i) = (red << 11) + (green << 5) + blue;
	}
	delete temp;
}

// ------------------------------------------------
// EOF